unit Unit1;

{ Lorentz's waterwheel simulator

  version 1.0   july 2022

  dimensions:
  1 pixel = 5mm
  wheel radius = 1 meter
  bucket shape is cilindrical,radius 10cm, height 30 cm
  bucket volume = 10 liter
  empty mass of total wheel = 15kg.
  friction 1N
  air friction 0.3N/v^2  (v is metres/sec)
  wood frame, plastic buckets

  time is in 1 millisec intervals assuming constant speed and water volume
  after 10 intervals the wheel picture is refreshed

  reporting is in
   - real time
   - slowed, doubled display intervals

  air friction F = 0.5 * rho * Cd * A * V^2
      F   : force in Newtons
      rho : air density = 1,2kg/m3
      Cd  : air drag coefficient = 0.3
      A   : area  (1 m^2)
      V   : speed m/sec

 version 1.1
   watersupply not restored at STOP . (fixed)
   waterjet * 1.5*abs(cos(pumpwheel angle))

 version 2.0
   15 buckets of 10 liters each
   lowered frame mass to 15kg
   lowered roll friction to 1N.
   Cd value set to 0.3

 version 3.0   aug. 2022
 increased braking
 max fill rate 20 l/sec.
}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, dav7components, StdCtrls, dav7RotationBtn, Buttons,
  shellapi;

type
  TForm1 = class(TForm)
    timer: Tdav7Timer;
    pump: TImage;
    tsplit: TImage;
    vopen: TImage;
    vclosed: TImage;
    glass: TImage;
    lefttop: TImage;
    righttop: TImage;
    leftbottom: TImage;
    rightbottom: TImage;
    inlet: TImage;
    hpipe1: TImage;
    hpipe2: TImage;
    hpipe3: TImage;
    vpipe1: TImage;
    vpipe2: TImage;
    nozzle: TImage;
    PaintBox1: TPaintBox;
    triplecon: TImage;
    longnozzle: TImage;
    hpipe4: TImage;
    PaintBox2: TPaintBox;
    filltext: TStaticText;
    leaktext: TStaticText;
    fillspeed1: TDav7RotationBtn;
    fillspeed2: TDav7RotationBtn;
    leakspeed1: TDav7RotationBtn;
    leakspeed2: TDav7RotationBtn;
    Label1: TLabel;
    Label2: TLabel;
    simtimetext: TStaticText;
    lightoff: TImage;
    lighton: TImage;
    SpeedButton1: TSpeedButton;
    SpeedButton2: TSpeedButton;
    PaintBox3: TPaintBox;
    proctimetext: TStaticText;
    Label4: TLabel;
    Label3: TLabel;
    leftBtn: TImage;
    rightBtn: TImage;
    Angletext: TStaticText;
    pauseBtn: TLabel;
    startBtn: TLabel;
    stopBtn: TLabel;
    msglabel: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    cycletext: TStaticText;
    modustext: TStaticText;
    Label7: TLabel;
    drop1: TImage;
    slowBtn: TSpeedButton;
    openhandle: TImage;
    closedhandle: TImage;
    helpBtn: TBitBtn;
    Label8: TLabel;
    bucket: TImage;
    frictionBtn: TDav7RotationBtn;
    Label9: TLabel;
    Label10: TLabel;
    frictionlabel: TStaticText;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure PaintBox1Paint(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure PaintBox2Paint(Sender: TObject);
    procedure fillspeed1Change(sender: TObject; position: Byte);
    procedure SpeedButton1Click(Sender: TObject);
    procedure SpeedButton2Click(Sender: TObject);
    procedure PaintBox3Paint(Sender: TObject);
    procedure pauseBtnMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure leftBtnMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure leftBtnMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure startBtnMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure slowBtnClick(Sender: TObject);
    procedure PaintBox1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure PaintBox1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure helpBtnClick(Sender: TObject);
    procedure frictionBtnChange(sender: TObject; position: Byte);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;
 
implementation

{$R *.dfm}

uses pgunit;

const pi2 : single = 2*pi;
      pi15 : single = 2*pi/15;
      davdata = 'DavData Software';
      rad2deg : single = 57.29578;
      deg2rad : single = 0.01745329;

//-- params

      emptymass   : single = 15;   // wheel + empty buckets (kg) (was 25)
      rolfriction : single = 1;    // Newton
      maxvolume = 10;              // litres per bucket
      interval = 10;               // millisecs

//-- valve controls

      valverect : array[1..3] of Trect = ((left:241;top:16;right:277;bottom:52),
                                          (left:291;top:16;right:327;bottom:52),
                                          (left:341;top:16;right:377;bottom:52));

type TBucket = record
                volume : single;
                height : byte; //in pixels
                posX : word;
                posY : word;
                vsin : single;
                vcos : single;
               end;
     TWheelstatus = (wsStopped,wsRunning,wsPaused,wsStopping);
     TControlmessage = (cmInitialise,cmStartBtnDown,cmPauseBtnDown,
                        cmStopBtnDown,cmStartBtnUp,cmPauseBtnUp,cmStopBtnUp,
                        cmWheelleft,cmWheelright,cmWheelStop,cmSpace);
     Trefstatus = (rsNone,rsXright,rsXLeft,rsUright,rsUleft);
     Trefpoint = record
                  status : Trefstatus;
                  angle  : single;
                  time   : int64;  //CPU time
                 end;

var map : Tbitmap;
    motormap : Tbitmap;
    speedmap : TBitmap;
    jetmap : Tbitmap;
    stopflag : boolean;
    bucket : array[1..15] of Tbucket;
    Awheel : single;       //waterwheel angle
    wheelspeed : single;   //in radians
    watersupply : single = 200;
    valveControl : byte = 2;  //bits 0: left 1:middle 2: right open
    valveselector : byte;     //1,2,3..
    dynfriction : single = 1;
//
    Amotorwheel : single;  //pump motor angle
    pumpspeed : single;
    fillspeed : single;
    leakspeed : single;
    simtime   : dword;     //total time in msecs.
    timeout   : dword;
//
    wheelstatus : TWheelstatus =wsStopped;
//
    oldcode : byte;//--0-:from ref,-1-:to ref;--0:right;--1:left
    refpoint : Trefpoint;//reference point to detect cycles
//
    volumeOffset : array[1..38] of byte;


//-- cycle time reporting

procedure initcyclereporting;
begin
 refpoint.status := rsNone;
 oldcode := 0;
end;

function makemovecode : byte;
//mc : movecode
//use reference point and wheelspeed
var DA : single;
begin
 if wheelspeed > 0 then result := $0 else result := $1;
//
 DA := refpoint.angle - Awheel;
 if DA >= pi then DA := DA - pi2;
 if DA < -pi then DA := DA + pi2;
//
 if ((wheelspeed > 0) and (DA > 0)) or
    ((wheelspeed < 0) and (DA < 0)) then result:= result or $2;
end;

procedure reportcycletime(cpt : int64);
var dd : double;
begin
 with form1 do
  begin
   dd := timer.CPUclockToTime(cpt)*1e-6;
   if slowBtn.Down then dd := dd*0.5;
   cycletext.caption := formatfloat('##0.###',dd);
  end;
end;

procedure processTurn(md : byte);
//md : move direction 0:right  1:left
var dt : int64;
begin
 with refpoint do
  begin
   angle := Awheel;
   dt := form1.timer.CPUtime;
   if ((md = 1) and (status = rsURight)) or
      ((md = 0) and (status = rsULeft)) then reportcycletime(dt - time)
    else form1.cycletext.Caption := '';
   form1.modustext.Caption := 'pendulum';
   time := dt;
   if md = 0 then status := rsURight else status := rsUleft;
  end;//with refpoint
end;

procedure processRotation(md : byte);
//by move buttons at top of form
//md : movedirection  0: right   1:left
var dt : int64;
begin
 with refpoint do
  begin
   if (((status = rsXRight) or (status = rsUright)) and (md = 0)) or
      (((status = rsXLeft)  or (status = rsUleft))  and (md = 1)) then
    begin
     dt := form1.timer.CPUtime;
     reportcycletime(dt-time);
     time := dt;
     if md=0 then status := rsXRight else status := rsXleft;
     form1.modustext.Caption := 'rotation';
    end;
  end;//with refpoint
end;

procedure referenceSetup;
begin
 with refpoint do
  begin
   angle := Awheel;
   time := form1.timer.CPUtime;
   if wheelspeed > 0 then
    begin
     status := rsXRight;
     oldcode := $0;
    end
    else begin
          status := rsXLeft;
          oldcode := $1;
         end;
  end;
end;

procedure CycleEvent;
var code : byte;
begin
 if wheelspeed = 0 then exit;
//
 if refpoint.status = rsNone then referenceSetup    //1st event
  else                                              //running
   begin
    code := makemovecode;
    if ((oldcode xor code) and $1) = $1 then        //if direction change
     begin
      processturn(code and $1);
      code := code and 1;         //force "from" reference point movement
     end
     else
      if ((oldcode and (code xor $2)) and $2) = $2 then
       processRotation(code and 1);
    oldcode := code;
   end;
end;

//-----------------------

procedure delay(usecs : dword);
//delay microseconds
//start time must be set before
begin
 with form1 do
  repeat
   application.ProcessMessages;
   timer.stop;
  until timer.elapsedtime > usecs;
end;

procedure presetVolumeOffsets;
//make array vlumeOffset
var i : byte;
begin
 for i := 1 to 38 do
  volumeOffset[i] := round(0.25*sqrt(400 - sqr(i-20)));
end;

function bucketheight(v : single) : byte;
//calculate height of water in bucket  in pixels.
//v : volume in liters
//maximum 60 pixels
begin
 result := round(v*6);
end;

procedure paintpumpsystem;
begin
 with map do with canvas do
  begin
   draw(320,540,form1.pump.picture.bitmap);
   if (Amotorwheel > 4.9) or (Amotorwheel < 1.8 ) then
    begin
     draw(310,595,form1.vopen.picture.bitmap); //left inlet valse
     draw(410,595,form1.vclosed.picture.bitmap);   //right inlet valve
     draw(310,490,form1.vclosed.picture.bitmap); //left outlet valve
     draw(410,490,form1.vopen.picture.bitmap);   //right outlet valve
    end
    else
     begin
      draw(310,595,form1.vclosed.picture.bitmap); //left inlet valve
      draw(410,595,form1.vopen.picture.bitmap);   //right inlet valve
      draw(310,490,form1.vopen.picture.bitmap);   //left outlet valve
      draw(410,490,form1.vclosed.picture.bitmap); //right outlet valve
     end;
   draw(400,455,form1.tsplit.picture.bitmap);
   draw(320,455,form1.lefttop.picture.bitmap);
   draw(540,435,form1.rightbottom.picture.bitmap);
   draw(540,0,form1.righttop.picture.bitmap);
   draw(545,340,form1.glass.picture.bitmap);
   draw(320,645,form1.inlet.picture.bitmap);
   draw(420,645,form1.inlet.picture.bitmap);
   draw(355,455,form1.hpipe1.picture.bitmap);
   draw(455,455,form1.hpipe2.picture.bitmap);
   draw(560,35,form1.vpipe1.picture.bitmap);
   draw(560,195,form1.vpipe2.picture.bitmap);
   draw(285,25,form1.nozzle.picture.bitmap);
   draw(235,25,form1.longnozzle.picture.bitmap);
   draw(335,25,form1.longnozzle.picture.bitmap);
   draw(220,0,form1.triplecon.Picture.Bitmap);
   draw(365,0,form1.hpipe4.picture.bitmap);
  end;
end;

procedure paintmap;
//background
var x,y : word;
begin
 with map do with canvas do
  begin
   brush.Color := $f0ffff;
   brush.Style := bsSolid;
   fillrect(rect(0,0,width,height));
   pen.Color := $fff0f0;
   pen.Width := 3;
   x := 0;
   while x < width do
    begin
     moveto(x,0);
     lineto(x,height);
     inc(x,40);
    end;
   y := 0;
   while y < height do
    begin
     moveto(0,y);
     lineto(width,y);
     inc(y,40);
    end;
   pen.Width := 1;               //water supply
   pen.Color := $ffc000;
   brush.Color := pen.Color;
   y := 740-round(watersupply*0.3);
   rectangle(240,y,520,740);
   pen.Width := 3;
   pen.Color := $000000;
   moveto(240,660);
   lineto(240,740);
   lineto(520,740);
   lineto(520,660);
  end;
end;

procedure paintmotor;
//in motor bitmap
var i,x : word;
begin
with motormap do with Canvas do
  begin
   brush.Color:= $ffffff;
   brush.Style := bsSolid;
   fillrect(rect(0,0,width,height));
   pen.width := 1;
   pen.color := $008000;
   for i := 0 to 165 do
    begin
     x := round(0.1*i);
     moveto(30-x,i+45);
     lineto(130+x,i+45);
    end;
   pen.color := $005000;
   pen.width := 3;
   brush.color:= $008000;
   brush.style := bsSolid;
   ellipse(30,5,130,105);
   moveto(30,45);
   lineto(13,198);
   lineto(146,198);
   lineto(130,45);
   pen.Color := $000060;
   pen.Width := 10;
   brush.Style := bsClear;
   pen.Width := 5;
   pen.color := $000060;
   brush.color := $000000;
   brush.style := bsSolid;
   font.Name := 'arial';
   font.Height := 14;
   font.Style := [fsBold];
   font.Color := $ffffff;
   brush.Style := bsClear;
   textout(35,170,davdata);
   font.Color := $000000;
   textout(36,171,davdata);
  end;
end;

procedure paintengine(a: single);
//a : angle in radians
var d,saveA,x,y : single;
    rx,ry : smallInt;
    i : byte;

  procedure joint(px,py,r : word);
  begin
   with map.canvas do
    begin
     pen.color := $f0f0f0;
     pen.width := 1;
     brush.color := $000000;
     brush.style := bsSolid;
     ellipse(px-r,py-r,px+r,py+r);
    end;
  end;

  procedure guideblock(px1,px2,y,h : word);
   begin
    with map.Canvas do
     begin
      brush.color := $606060;
      pen.Color := $d0d0d0;
      pen.Width:= 1;
      brush.Style := bsSolid;
      rectangle(px1,y,px2,y+h);
     end;
   end;
begin
 saveA := a;
 with map.Canvas do
  begin
   guideblock(180,300,563,10);
   guideblock(180,300,578,10);
   pen.width := 1;
   pen.color := $008000;
   pen.Color := $000060;
   pen.Width := 10;
   brush.Style := bsClear;
   ellipse(80,535,160,615);   //wheel 80 diam
   pen.width := 5;
   for i := 0 to 3 do         //spokes
    begin
     moveto(120,575);
     rx := round(120+40*sin(a));
     ry := round(575-40*cos(a));
     moveto(120,575);
     lineto(rx,ry);
     a := a + 0.5*pi;
    end;
    pen.Width := 5;
    pen.color := $000060;
    brush.color := $000000;
    brush.style := bsSolid;
    ellipse(110,565,130,585);  //shaft
    x := 40*sin(saveA);
    y := 40*cos(saveA);
    rx := round(120+x);
    ry := round(575-y);
    joint(rx,ry,7);
    d := x+sqrt(14400-sqr(y));
    pen.Width := 5;
    pen.color := $000000;
    moveto(rx,ry);
    rx := round(120+d);
    lineto(rx,575);
    pen.Width := 1;
    joint(rx,575,7);
    rx := round(d+260);
    pen.width := 5;
    pen.color := $000000;
    lineto(rx,575);
    pen.Width := 1;
    pen.Color := $000060;
    for i := 0 to 9 do           //piston
     begin
      moveto(rx+i,563);
      lineto(rx+i,587);
     end;
    pen.color := $e0e0e0;
    moveto(rx,563);
    lineto(rx+10,563);
    moveto(rx,586);
    lineto(rx+10,586);
  end;
end;

procedure paintbucketvolume(n : byte);
//paint volume in bucket n
//6 pix / liter
var x,y1,y2 : word;
    h,i : byte;
begin
 with bucket[n] do
  begin
   h := height;
   x := bucket[n].posX-20;
   y2 := bucket[n].posY + 65;
   y1 := y2 - h;
  end;
 with map.Canvas do
  begin
   pen.Width := 1;
   pen.color := $ff0000;
   for i := 1 to 39 do
    begin
     moveto(x + i,y1 + volumeOffset[i]);
     lineto(x + i,y2 + volumeOffset[i]);
    end;
  end;
end;

procedure placebuckets;
var dropcount,i,j : byte;
begin
 for i := 1 to 15 do
  begin
   with bucket[i] do
    begin
     map.Canvas.draw(posX-25,posY-5,form1.bucket.Picture.Bitmap);
     dropcount :=round(leakspeed*height+14) div 15;
     if dropcount > 3 then dropcount := 3;
     for j := 1 to dropcount do
      map.Canvas.Draw(posX-5,posY+60+j*12,form1.drop1.picture.bitmap);
    end;
  end;
end;

procedure paintwheel(a : single);
var i : byte;
    aixpi15 : single;
begin
 for i := 1 to 15 do             //set bucket coordinates
  with bucket[i] do
   begin
    aixpi15 := a+(i-1)*pi15;
    vcos := cos(aixpi15);
    vsin := sin(aixpi15);
    posX := 300 + round(200*vsin);
    posY := 280 - round(200*vcos);
   end;
//
 with map.Canvas do              //back part of wheel
  begin
   pen.Color := $000080;
   pen.Width := 5;
   moveto(bucket[1].posX,bucket[1].posY);
   for i := 2 to 15 do lineto(bucket[i].posX,bucket[i].posY);
   lineto(bucket[1].posX,bucket[1].posY);
   for i := 1 to 15 do
    begin
     if i = 1 then pen.color := $0060ff else pen.color := $000080;
     moveto(300,280); lineto(bucket[i].posX,bucket[i].posY); //back spoke
    end;
   pen.Width := 1;
   pen.Color := $000060;
   brush.Color := $000060;
   ellipse(280,260,320,300);
   pen.color := $000000;
   brush.Color := $000000;
   ellipse(290,270,310,290);
//
   placebuckets;
   for i := 1 to 15 do
    if bucket[i].height > 0 then paintbucketVolume(i);
//
   pen.width := 5;             //front wheel
   pen.color := $0000a0;
   moveto(bucket[1].posX,bucket[1].posY+15);
   for i := 2 to 15 do lineto(bucket[i].posX,bucket[i].posY+15);
   lineto(bucket[1].posX,bucket[1].posY+15);
  end;
end;

procedure paintblueArrow(x,y : word);
//for fill/leak bucket
begin
 with form1.Canvas do
  begin
   pen.Color := $ff0000;
   pen.Width := 2;
   moveto(x,y);
   lineto(x,y+30);
   lineto(x-5,y+20);
   moveto(x,y+30);
   lineto(x+5,y+20);
  end;
end;

procedure drawmotorlight(lon : boolean);
begin
 with form1 do
 if lon then map.Canvas.Draw(100,640,lighton.Picture.bitmap)
  else map.Canvas.Draw(100,640,lightoff.Picture.bitmap);
end;

procedure paintspeedmap;
//speed needle indicator
var i : byte;
    x1,y1,x2,y2 : word;
    a,sina,cosa : single;
    s : string;
begin
 with speedmap do with canvas do
  begin
   brush.Color := $f0ffff;
   brush.Style := bsSolid;
   fillrect(rect(0,0,width,height));
   pen.Color := $000000;
   brush.Style := bsSolid;
   pen.Width := 1;
   pen.color := $000000;
   brush.Color := $000000;
   ellipse(84,89,97,102);
   pen.width := 2;
   font.name := 'arial';
   font.style := [fsBold];
   font.height := 14;
   font.color := $000000;
   brush.style := bsClear;
   for i := 0 to 10 do
    begin
     a := (30+i*12)*deg2rad;
     sina := sin(a);
     cosa := cos(a);
     x1 := round(90-75*cosa);
     x2 := round(90-80*cosa);
     y1 := round(95-75*sina);
     y2 := round(95-80*sina);
     if i > 0 then lineto(x1,y1) else moveto(x1,y1);
     lineto(x2,y2);
     moveto(x1,y1);
    end;
   for i := 0 to 5 do
    begin
     a := (30+i*24)*deg2rad;
     sina := sin(a);
     cosa := cos(a);
     s := inttostr(i);
     x1 := round(90-90*cosa-i);
     y1 := round(95-95*sina);
     textout(x1,y1,s);
    end;
   font.Height := 12;
   font.Color := $0000a0;
   textout(0,0,'speed');
   textout(0,10,'(m/sec)');
  end;
end;

procedure paintspeed(speed : single);
var a : single;
    x1,y1 : word;
    sina,cosa : single;
    i : byte;
begin
 speed := abs(speed);
 with form1.PaintBox3.canvas do
  begin
   draw(0,0,speedmap);
   a := (30+speed*24)*deg2rad;
   sina := sin(a);
   cosa := cos(a);
   x1 := round(90-75*cosa);
   y1 := round(95-75*sina);
   pen.Color := $0000a0;
   pen.Width := 3;
   for I := 0 to 1 do
    begin
     moveto(90,95);
     lineto(x1,y1);
     pen.Color := $00ffff;
     pen.Width := 1;
   end;
  end;
end;

procedure placemotor;
begin
 map.canvas.draw(40,520,motormap);
end;

procedure showsimtime;
//elapsed time clock
//time is in msecs intervals
const m : array[1..7] of byte = (10,10,6,10,6,10,6);
      zero = ord('0');
var ta : array[1..7] of byte;
    t : dword;
    s : string;
    i : byte;
begin
 t := simtime div 100; //make 10th of seconds
 s := '';
 for i := 1 to 7 do
  begin
   ta[i] := t mod m[i];
   t := t div m[i];
  end;
 s := char(zero + ta[7]) + char(zero + ta[6]) + ':';
 s := s + char(zero + ta[5]) + char(zero + ta[4]) +':' ;
 s := s + char(zero + ta[3]) + char(zero + ta[2])+ ':' ;
 s := s + char(zero + ta[1]);
 form1.simtimetext.caption := s;
end;

procedure showfillleakspeed;
const fc = '#0.0';
begin
 with form1 do
  begin
   filltext.Caption := formatfloat(fc,fillspeed);
   leaktext.Caption := formatfloat(fc,leakspeed);
  end;
end;

procedure showWheelPos;
//display angle position of waterwheel
var angle : smallInt;
begin
 with form1.Angletext do
  begin
   angle := round(Awheel*rad2deg);
   if angle < 180 then
    begin
     font.color := $000000;
     caption := inttostr(angle);
    end
    else begin
          font.color := $0000ff;
          caption := inttostr(angle - 360);
         end;
  end;
end;

procedure paintControlEdges(ctrl : TControl; color1,color2 : dword; w : byte);
var x1,y1,x2,y2 : word;
    i : byte;
begin
 with ctrl do
  begin
   x1 := Left-w;
   y1 := Top-w;
   dec(w);
   x2 := left+width+w;
   y2 := top+height+w;
  end;
 with form1.canvas do
  begin
   pen.width := 1;
   for i := 0 to w-1 do
    begin
     pen.Color := color1;
     moveto(x2-i,y1+i);
     lineto(x1+i,y1+i);
     lineto(x1+i,y2-i);
     pen.Color := color2;
     lineto(x2-i,y2-i);
     lineto(x2-i,y1+i);
    end;
  end;
end;

procedure paintnozzlevalves;
const vx : array[0..2] of word = (241,291,341);
var i : byte;
begin
 with form1 do
  for i := 0 to 2 do
   if ((1 shl i) and valvecontrol) > 0 then
    map.canvas.draw(vx[i],16,openhandle.picture.Bitmap)
    else map.canvas.draw(vx[i],36,closedhandle.picture.Bitmap);
end;

procedure paintjetmap;
const posx : array[0..2] of word = (247,297,347);
      posy : array[0..2] of word = (72,62,72);
type PDW = ^dword;
var p0,py : dword;
    i,x,y : byte;
    stream : byte;
begin
 randomize;
 stream := round(10 + 4*pumpspeed); //10..90
 with jetmap do with canvas do
  begin
   brush.color := $ff0000;
   brush.Style := bsSolid;
   fillrect(rect(0,0,width,height));
   p0 := dword(scanline[0]);
   for y := 0 to height-1 do
    begin
     py := p0 - 28*y;
     for x := 0 to width-1 do
      if random(100) > stream then PDW(py + (x shl 2))^ := $ffffff;//make white
    end;
  end;
 if pumpspeed > 0 then
  for i := 0 to 2 do
   if (valvecontrol and (1 shl i)) > 0 then map.Canvas.Draw(posx[i],posy[i],jetmap);
end;

procedure paintvalvecontrols;
//paint rectangle around selected valse
//paint in paintbox1
begin
 if valveselector = 0 then exit;
//
 with form1.paintbox1.canvas do
  begin
   pen.width := 2;
   pen.color := $ff0000;
   brush.Style := bsClear;
   rectangle(valverect[valveselector]);
  end;
end;

procedure clearvalveselection;
const r : Trect = (left:238; top:15; right:380;bottom:55);
//called by mouse move event
begin
 form1.PaintBox1.Canvas.CopyRect(r,map.Canvas,r);
end;

procedure showcontrolbuttons(cc : byte);
//bits 0,1,2  show buttons start,pause,stop   0:up  1:down
var i : byte;
    color1,color2 : dword;
    LBL : TLabel;
begin
 with form1 do
  for i := 0 to 2 do
   begin
    if ((1 shl i) and cc) <> 0 then color1 := $000000 else color1 := $ffffff;
    color2 := color1 xor $ffffff;
    case i of
     0 : LBL := startBtn;
     1 : LBL := pauseBtn;
     2 : LBL := stopBtn;
     else LBL := nil;
    end;//case
    paintcontrolEdges(LBL,color1,color2,5);
   end;
end;

procedure repaint;
begin
 paintmap;
 paintwheel(Awheel);
 placemotor;
 paintpumpsystem;
 paintnozzlevalves;
 paintjetmap;
 paintengine(Amotorwheel);
 drawmotorlight(pumpspeed > 0);
 form1.PaintBox1.Canvas.Draw(0,0,map);
 paintvalvecontrols;
 paintspeed(wheelspeed);
end;

procedure simulate;
//time slice is 1 msec.
//process wheel speed, wheel position and bucket volumes
var wdeg : smallInt;
    i : byte;
    drop,dv,dForce,force,oldmass,newmass,leak,overflow : single;
    fillup : boolean;
begin
 force := 0;
 leak := 0;
 overflow := 0;
 oldmass := emptymass;
 newmass := emptymass;
 for i := 1 to 15 do        //for all buckets
  with bucket[i] do
   begin
    oldmass := oldmass + volume;
    fillup := false;
    wdeg := round((Awheel+(i-1)*pi15)*rad2deg);
    if wdeg >= 360 then wdeg := wdeg - 360;
    if (valvecontrol and $1) > 0 then
     fillup := fillup or ((wdeg > 340) and (wdeg <= 352));
    if (valvecontrol and $2) > 0 then
     fillup := fillup or ((wdeg >= 354) or (wdeg < 7));
    if (valvecontrol and $4) > 0 then
     fillup := fillup or ((wdeg >= 8) and (wdeg < 20));
//
    if fillup then
     begin
      dv := 0.0015*pumpspeed*abs(cos(Amotorwheel));
      volume := volume + dv;
      if volume > maxvolume then begin
                           overflow := overflow + (volume-maxvolume);
                           volume := maxvolume;
                          end;
     end;
    drop := 0.001*leakspeed;
    leak := leak + drop;
    volume := volume - drop;
    if volume < 0 then volume := 0;
    newmass := newmass + volume;
    height := bucketheight(volume);
    force := force+9.8*volume*sin(Awheel+(i-1)*pi15);
   end;//with bucket[ ]
//
 if (abs(force) <= rolfriction + 0.1) and (abs(wheelspeed) < 1e-3) and
    (pumpspeed > 0) and (valvecontrol > 0) then
  begin
   dec(timeout);
   if timeout = 0 then
    begin
     wdeg := round(Awheel*rad2deg);
     wdeg := wdeg mod 24;
     if wdeg < 12 then Awheel := Awheel - deg2rad
      else Awheel := Awheel + deg2rad;
     if Awheel < 0 then Awheel := Awheel + pi2;
     if Awheel >= pi2 then Awheel := Awheel - pi2;
     timeout := 5000;
   end;
  end else timeout := 5000;
//
 if abs(wheelspeed) > 0.001 then
  wheelspeed := wheelspeed*sqrt((oldmass-leak-overflow)/(newmass+overflow));
 Awheel := Awheel + 0.001*wheelspeed;
 if Awheel > pi2 then Awheel := Awheel - pi2
  else if Awheel < 0 then Awheel := Awheel + pi2;
 dForce := rolfriction + dynfriction*sqr(wheelspeed);   //roll + air friction == ??
 if wheelspeed > 0 then force := force-dForce
  else if wheelspeed < 0 then force := force+dForce;
 wheelspeed := wheelspeed + 0.001*force/newmass;
//
 Amotorwheel := Amotorwheel + 0.001*pumpspeed; //==
 if Amotorwheel > pi2 then Amotorwheel := Amotorwheel - pi2;
 watersupply := 200-(newmass-emptymass);
 simreport(Awheel,Wheelspeed); //--> pgunit
 CycleEvent;                   //motion type and cycle time reporting
end;

procedure rotate;
var n : word;
    xx : byte;
    delaytime : dword;
begin
 xx := 0;
 stopflag := false;
 timeout := 5000; //millisecs
 repeat
   form1.timer.start;
   for n := 1 to interval do simulate;

//-- show results

   if pumpspeed < fillspeed then pumpspeed := pumpspeed + 0.05; //accellerate
   if pumpspeed > fillspeed then pumpspeed := pumpspeed - 0.05; //decellerate
   if abs(pumpspeed) < 0.05 then pumpspeed := 0;
   repaint;          //wheel,buckets,pump system
   showpgmap;
   pgTimeErase;
   inc(simtime,interval);    // simulator clock
   showsimtime;
   form1.timer.stop;
   inc(xx);
   if xx > 40 then
    begin
     form1.proctimetext.Caption := formatfloat('#0.00',form1.timer.elapsedtime/(10*interval));
     xx := 0;
    end;
   delaytime := interval*1000; 
   if form1.slowBtn.down then delaytime := delaytime shl 1;    //*2
   delay(delaytime);  //includes windows message processing
 until stopflag;
end;

procedure procwheelStop;
var n : byte;
    mass : single;
    delaytime : dword;
begin
 stopflag := true;
 repeat
  mass := 0;
  form1.timer.start;
  for n :=1 to 15 do
   with bucket[n] do
    begin
     volume := volume - 0.1;
     if volume < 0.1 then volume := 0;
     height := bucketHeight(volume);
     mass := mass + volume;
    end;
  Awheel := Awheel + 0.01*wheelspeed;
  if Awheel >= pi2 then Awheel := Awheel - pi2
   else if Awheel < 0 then Awheel := Awheel + pi2;
  if wheelspeed >= 0 then begin
                           wheelspeed := wheelspeed - 0.01;
                           if wheelspeed < 0 then wheelspeed := 0;
                          end
   else begin
         wheelspeed := wheelspeed + 0.01;
         if wheelspeed > 0 then wheelspeed := 0;
        end;
  watersupply := 200 - mass;
  repaint;
  delaytime := 10000;
  if form1.slowBtn.Down then delaytime := delaytime shl 1;//*2
  delay(delaytime);
 until (wheelspeed = 0) and (mass = 0)
end;

procedure resetcontrols;
begin
 Awheel := deg2rad; //1 degree right
 form1.Angletext.caption := '1';
 Amotorwheel := 0;
 Wheelspeed := 0;
 fillspeed := 0;
 leakspeed := 0;
 with form1 do
  begin
   fillspeed1.position := 0;
   fillspeed2.position := 0;
   leakspeed1.position := 0;
   leakspeed2.position := 0;
   showfillleakspeed;
  end;
end;

procedure startWheel;
begin
 simtime := 0;
 showsimtime;
 initcyclereporting;
 form1.cycletext.caption := '';
 form1.modustext.Caption := '';
 form1.Angletext.Caption := '';
 form1.msglabel.Caption := 'running';
 showcontrolbuttons(1);
 pumpspeed := 0;
 wheelstatus := wsRunning;
 form1.angletext.caption := '';
 paintpolargraphmap;
 drawpgmap;
 rotate;
end;

procedure pauseStop;
begin
 form1.msglabel.Caption := 'paused';
 stopflag := true;
 wheelstatus := wsPaused;
 showcontrolbuttons(3);
end;

procedure pauseGO;
begin
 form1.msglabel.caption := 'running';
 wheelstatus := wsRunning;
 showcontrolbuttons(1);
 rotate;
end;

procedure stopWheel;
var deg : smallInt;
begin
 form1.msglabel.Caption := 'stopping';
 showcontrolbuttons(4);
 pumpspeed := 0;
 wheelstatus := wsStopping;
 procWheelstop;
 wheelstatus := wsStopped;
 form1.msglabel.caption := 'stopped';
 deg := round(Awheel*rad2deg);
 if deg >= 180 then deg := deg - 360;
 form1.angletext.caption := inttostr(deg);
end;

procedure movewheel(drctn : shortInt);
//drctn : degrees per step
var delaytime : dword;
    deg : smallInt;
begin
 stopflag := false;
 delaytime := 330000;
 repeat
  AWheel := Awheel + drctn*deg2rad;
  deg := round(AWheel*rad2deg);
  if deg >= 180 then deg := deg - 360;
  if deg < -180 then deg := deg + 360;
  form1.Angletext.Caption := inttostr(deg);
  form1.timer.start;
  delay(delaytime);
  if delaytime > 30000 then delaytime := delaytime - 10000;
  repaint;
 until stopflag;
end;

procedure InitProc;
begin
 resetcontrols;
 wheelstatus := wsStopped;
 repaint;
 simtime := 0;
 showsimtime;
end;

procedure moveWheelleft;
begin
 with form1 do paintControlEdges(leftBtn,$000000,$ffffff,5);
 moveWheel(-1);
end;

procedure movewheelright;
begin
 with form1 do paintControlEdges(rightBtn,$000000,$ffffff,5);
 movewheel(1);
end;

procedure stopwheelmotion;
begin
 with form1 do
  begin
   paintControlEdges(leftBtn,$ffffff,$000000,5);
   paintControlEdges(rightBtn,$ffffff,$000000,5);
  end; 
 stopflag := true;
end;

procedure controlcenter(controlmessage : Tcontrolmessage);
begin
 case controlmessage of
  cmInitialise   : initproc;
  cmStartBtnDown : if wheelstatus = wsStopped then startWheel;
  cmPauseBtnDown : if wheelstatus = wsRunning then pauseStop
                    else if wheelstatus = wsPaused then pauseGO;
  cmStopBtnDown  : if wheelstatus = wsRunning then stopwheel;
  cmWheelLeft    : if wheelstatus = wsStopped then movewheelleft;
  cmWheelRight   : if wheelstatus = wsStopped then movewheelright;
  cmWheelStop    : if wheelstatus = wsStopped then stopWheelmotion;
 end;
end;

procedure Setpixelformat32;
//purpose is speed increase for draw method
begin
 with form1 do
  begin
   glass.picture.bitmap.pixelformat := pf32bit;
   vopen.picture.bitmap.pixelformat := pf32bit;
   vclosed.picture.bitmap.pixelformat := pf32bit;
   tsplit.picture.bitmap.pixelformat := pf32bit;
   lefttop.picture.bitmap.pixelformat := pf32bit;
   rightbottom.picture.bitmap.pixelformat := pf32bit;
   righttop.picture.bitmap.pixelformat := pf32bit;
   inlet.picture.bitmap.pixelformat := pf32bit;
   hpipe1.picture.bitmap.pixelformat := pf32bit;
   hpipe2.picture.bitmap.pixelformat := pf32bit;
   hpipe3.picture.bitmap.pixelformat := pf32bit;
   vpipe1.picture.bitmap.pixelformat := pf32bit;
   vpipe2.picture.bitmap.pixelformat := pf32bit;
   nozzle.picture.bitmap.pixelformat := pf32bit;
   bucket.picture.Bitmap.PixelFormat := pf32bit;
  end;
end;

procedure createProc;
//create maps
begin
 map := TBitmap.Create;
 with map do
  begin
   pixelformat := pf32bit;
   width := form1.paintbox1.width;
   height := form1.paintbox1.height;
  end;
 motormap := TBitmap.create;
 with motormap do
  begin
   pixelformat := pf32bit;
   transparent := true;
   transparentcolor := $ffffff;
   width := 160;
   height := 200;
  end;
 speedmap :=Tbitmap.Create;
 with speedmap do
  begin
   pixelformat := pf32bit;
   width := form1.paintbox3.Width;
   height := form1.paintbox3.Height;
  end;
 jetmap := Tbitmap.Create;
 with jetmap do
  begin
   width := 7;
   height := 30;
   pixelformat := pf32bit;
   transparent := true;
   transparentcolor := $ffffff;
  end;
//
 createPGmap;
 paintmotor;
 paintpolargraphmap;
 paintspeedmap;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
 createProc;
 setpixelformat32;
 presetVolumeOffsets;
 activecontrol := nil;
 controlcenter(cmInitialise);
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
 map.Free;
 speedmap.Free;
 motormap.Free;
 jetmap.Free;
 destroyPGmap;
end;

procedure TForm1.PaintBox1Paint(Sender: TObject);
//Lorentz wheel
begin
 paintbox1.canvas.draw(0,0,map);
end;

procedure TForm1.FormPaint(Sender: TObject);
begin
 paintControledges(paintbox1,$000000,$808080,3);
 paintControlEdges(paintbox2,$000000,$808080,3);
 paintControlEdges(paintbox3,$000000,$808080,3);
 canvas.Draw(1000,600,bucket.Picture.Bitmap);//===
 paintblueArrow(1025,580);
 paintblueArrow(1025,680);
 paintcontroledges(startBtn,$f0f0f0,$000000,5);
 paintcontroledges(pauseBtn,$f0f0f0,$000000,5);
 paintcontroledges(stopBtn,$f0f0f0,$000000,5);
 paintcontrolEdges(leftBtn,$f0f0f0,$000000,5);
 paintcontrolEdges(rightBtn,$f0f0f0,$000000,5);
 paintcontrolEdges(msglabel,$00a0ff,$00c0ff,3);
end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
 stopflag := true;
end;

procedure TForm1.PaintBox2Paint(Sender: TObject);
//polargraph
begin
 drawPGmap;
end;

procedure TForm1.fillspeed1Change(sender: TObject; position: Byte);
//tag 1..4 selects button
var btn : TDav7RotationBtn;
begin
 btn := TDav7RotationBtn(sender);
 case btn.tag of
   1 : begin
        fillspeed := fillspeed1.position;
        if fillspeed2.position > 0 then fillspeed2.position := 0;
       end;
   2 : begin
        fillspeed := fillspeed1.position + fillspeed2.position*0.1;
       end;
   3 : begin
        leakspeed := leakspeed1.position;
        if leakspeed2.position > 0 then leakspeed2.position := 0;
       end;
   4 : begin
        leakspeed := leakspeed1.position + leakspeed2.position*0.1;
       end;
 end;
 showfillleakspeed;
end;

procedure TForm1.SpeedButton1Click(Sender: TObject);
//clear clock
begin
 simtime := 0;
 showsimtime;
end;

procedure TForm1.SpeedButton2Click(Sender: TObject);
//repaint graphmap
begin
 paintpolargraphmap;
 drawPGmap;
end;

procedure TForm1.PaintBox3Paint(Sender: TObject);
begin
 paintspeed(wheelSpeed);
end;

procedure TForm1.pauseBtnMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
 case TLabel(sender).Tag of
  1 : controlcenter(cmStartBtnDown);
  2 : controlcenter(cmpauseBtnDown);
  3 : controlcenter(cmStopBtnDown);
 end;
end;

procedure TForm1.startBtnMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
 case Tcontrol(sender).tag of
  1 : controlcenter(cmStartBtnUp);
  2 : controlcenter(cmpauseBtnUp);
  3 : controlcenter(cmStopBtnUp);
 end;
end;

procedure TForm1.leftBtnMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
 case TImage(sender).Tag of
  1 : controlCenter(cmWheelLeft);
  2 : controlcenter(cmWheelRight);
 end;
end;

procedure TForm1.leftBtnMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
 controlcenter(cmWheelStop);
end;

procedure TForm1.slowBtnClick(Sender: TObject);
begin
 cycletext.Caption := '';
end;

procedure TForm1.PaintBox1MouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
//valve control
var i,vn : byte;
    hit : boolean;
begin
 hit :=false;
 i := 0;
 vn := 0;
 while (hit = false) and (i < 3) do
  begin
   inc(i);
   with valverect[i] do
    hit := (x > left) and (x < right) and (y > top) and (y <bottom);
  end;//while
  if hit then vn := i;
  if vn <> valveselector then
   begin
    if valveselector > 0 then clearvalveselection;
    valveselector := vn;
    if (valveselector > 0) and (wheelstatus <> wsRunning) then paintvalvecontrols;
   end;
end;

procedure TForm1.PaintBox1MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
//for valve on/off control
begin
 if valveselector > 0 then
  begin
   valvecontrol := valvecontrol xor (1 shl (valveselector-1));
   if (wheelstatus <> wsRunning) then unit1.repaint;
  end;
end;

procedure TForm1.helpBtnClick(Sender: TObject);
//website
begin
 activecontrol := nil;
 ShellExecute(0,'open','http://www.davdata.nl/math/waterwheel.html', nil, nil, SW_SHOWNORMAL);
end;

procedure TForm1.frictionBtnChange(sender: TObject; position: Byte);
begin
 dynfriction := frictionBtn.position*0.1;
 frictionlabel.Caption := formatfloat('0.0',dynfriction);
end;

end.
